/*
 * Copyright (C) 2013-2016 CERN (www.cern.ch)
 * Author: Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */

#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include "mockturtle-rt.h"


/**
 * It sends a character to the UART interface
 * @param[in] c the character to sent
 * @return 0 on success (for the time being it does not fail)
 */
int putchar(int c)
{
	lr_writel(c, MT_CPU_LR_REG_UART_CHR);

	return 0;
}


/**
 * It sends a string over the serial interface
 * @param[in] p string to send
 * @return number of sent characters
 *
 * The function does not add any special character.
 * If you neeed the new-line or the carriage-return
 * you have to add them to your strings.
 */
int puts(const char *p)
{
	const char *p_orig = p;
	char c;

	while ((c = *(p++)))
		putchar(c);

	return p - p_orig;
}


/**
 * It prints on the serial console the given message
 * @param[in] msg a mock turtle message
 */
void pr_message(struct trtl_fw_msg *msg)
{
	struct trtl_hmq_header *h = msg->header;
	uint32_t *d = msg->payload;
	int i;

	if (!HAS_MOCKTURTLE_LIBRARY_PRINT_MESSAGE_ENABLE)
		return;

	pp_printf("Message\n\r");
	pp_printf("\tapp_id 0x%04"PRIx16" | flags   0x%02"PRIx8" | msg_id 0x%02"PRIx8"\r\n",
		 h->rt_app_id, h->flags, h->msg_id);
	pp_printf("\tlen 0x%04"PRIx16" | sync_id 0x%04"PRIx16"\r\n",
		 h->len, h->sync_id);
	pp_printf("\tseq 0x%08"PRIx32"\r\n",
		 h->seq);

	for (i = 0; i < h->len; i++)
		pp_printf("%s: data[%d] = 0x%"PRIx32"\n\r", __func__, i , d[i]);
}

static inline void rmq_ep_out_writel( int slot, uint32_t data, uint32_t addr )
{
  mq_writel( TRTL_RMQ, data, addr + TRTL_MQ_SLOT_ENDPOINT_CONFIG_START + TRTL_MQ_SLOT_OUT (slot) );
}

static inline void rmq_ep_in_writel( int slot, uint32_t data, uint32_t addr )
{
  mq_writel( TRTL_RMQ, data, addr + TRTL_MQ_SLOT_ENDPOINT_CONFIG_START + TRTL_MQ_SLOT_IN (slot) );
}

static void rmq_ep_bind_in ( int slot, struct trtl_ep_address* addr )
{
  uint32_t cfg = 0;

  switch(addr->type)
  {
    case TRTL_EP_FRAME_UDP:
      cfg |= TRTL_EP_FILTER_UDP;
      break;
    case TRTL_EP_FRAME_TLV:
      cfg |= TRTL_EP_FILTER_UDP;
      break;
    case TRTL_EP_FRAME_RAW:
      cfg |= TRTL_EP_FILTER_RAW;
      break;
  }

  cfg |= addr->filter;
  uint32_t tmp;
  rmq_ep_in_writel( slot, cfg, TRTL_EP_IN_CONFIG );
  tmp = ( (uint32_t)addr->dst_mac[0] << 8 ) | addr->dst_mac[1];

  rmq_ep_in_writel( slot, tmp, TRTL_EP_IN_DST_MAC_HI );
  tmp = ( (uint32_t)addr->dst_mac[2] << 24 ) |
        ( (uint32_t)addr->dst_mac[3] << 16 ) |
        ( (uint32_t)addr->dst_mac[4] << 8 ) |
        ( (uint32_t)addr->dst_mac[5] << 0 );

  rmq_ep_in_writel( slot, tmp, TRTL_EP_IN_DST_MAC_LO );
  rmq_ep_in_writel( slot, addr->ethertype, TRTL_EP_IN_ETHERTYPE );
  rmq_ep_in_writel( slot, addr->dst_ip, TRTL_EP_IN_DST_IP );
  rmq_ep_in_writel( slot, addr->dst_port, TRTL_EP_IN_DST_PORT );
}

static void rmq_ep_bind_out ( int slot, struct trtl_ep_address* addr )
{
  uint32_t tmp;

  rmq_ep_out_writel( slot, addr->type == TRTL_EP_FRAME_UDP ? 1 : 0, TRTL_EP_OUT_CONFIG );

  tmp = ( (uint32_t)addr->dst_mac[0] << 8 ) | addr->dst_mac[1];
  rmq_ep_out_writel( slot, tmp, TRTL_EP_OUT_MAC_HI );

  tmp = ( (uint32_t)addr->dst_mac[2] << 24 ) |
        ( (uint32_t)addr->dst_mac[3] << 16 ) |
        ( (uint32_t)addr->dst_mac[4] << 8 ) |
        ( (uint32_t)addr->dst_mac[5] << 0 );

  rmq_ep_out_writel( slot, tmp, TRTL_EP_OUT_MAC_LO );
  rmq_ep_out_writel( slot, addr->ethertype, TRTL_EP_OUT_ETHERTYPE );
  rmq_ep_out_writel( slot, addr->src_ip, TRTL_EP_OUT_SRC_IP );
  rmq_ep_out_writel( slot, addr->dst_ip, TRTL_EP_OUT_DST_IP );
  rmq_ep_out_writel( slot, addr->src_port, TRTL_EP_OUT_SRC_PORT );
  rmq_ep_out_writel( slot, addr->dst_port, TRTL_EP_OUT_DST_PORT );


}

void mq_bind(enum trtl_mq_type type, int slot, int is_out, struct trtl_ep_address* addr)
{
  if( type != TRTL_RMQ )
  {
    pr_error("mq_bind() not supported on HMQ slots\n\r");
    return;
  }

  if(is_out)
    rmq_ep_bind_out( slot, addr );
  else
    rmq_ep_bind_in( slot, addr );
}
