#include <mockturtle-rt.h>

// maximum payload size of the test packets, in 32-bit words
#define MAX_RMQ_PAYLOAD 128

// maximum header (UDP+IP+Ethernet) size of the test packets, in 32-bit words
#define MAX_RMQ_HEADER 128

#define TEST_MESSAGE_ID 0xdeadbeef

struct trtl_test_message
{
	// Message id (0xdeadbeef)
	uint32_t id;
	// MT CPU that sent the message
	uint32_t cpu;
	// Sequence number
	uint32_t seq;
	// Statistics: total number of packets received by the MT CPU
	uint32_t n_rx;
	// Statistics: number of packets missed (by finding gaps in the sequential numbers)
	uint32_t n_missed;
	// Statistics: number of packets received with errors
	uint32_t n_errors;
	// Payload size in 32-bit words
	uint32_t payload_len;
	uint32_t payload[MAX_RMQ_PAYLOAD];
};

uint32_t tx_seq_count = 0;
uint32_t rx_count = 0;
uint32_t rx_last_id = -1;
uint32_t rx_missed = 0;
uint32_t rx_errors = 0;

void send_reply(int cpu_id, int rmq, int payload_len)
{
	struct trtl_fw_msg msg;

	// queue full? wait
	if (!mq_claim(TRTL_RMQ, rmq))
		return;

	mq_map_out_message(TRTL_RMQ, rmq, &msg);

	msg.header->len = payload_len + 7;
	struct trtl_test_message *p = msg.payload;

	p->id = TEST_MESSAGE_ID;
	p->cpu = cpu_id;
	p->seq = tx_seq_count++;
	p->n_rx = rx_count;
	p->n_missed = rx_missed;
	p->n_errors = rx_errors;
	p->payload_len = payload_len;

	// make some dummy payload
	for (int i = 0; i < payload_len; i++)
	{
		p->payload[i] = (((i << 16) | i) ^ 0xffff0000);
	}

	mq_send(TRTL_RMQ, rmq);
}

void handle_rx(int cpu_id, int rmq)
{
	if (!mq_poll_in(TRTL_RMQ, 1 << rmq))
		return;

	struct trtl_fw_msg tmsg;

	mq_map_in_message(TRTL_RMQ, rmq, &tmsg);

	struct trtl_test_message *p = tmsg.payload;

	if (p->id != TEST_MESSAGE_ID)
	{
		rx_errors++;
		mq_discard(TRTL_RMQ, rmq);
		return;
	}

	// validate the payload
	for (int i = 0; i < p->payload_len; i++)
	{
		if (p->payload[i] != (((i << 16) | i) ^ 0xffff0000))
		{
			rx_errors++;
			break;
		}
	}

	// look for missing packets
	if (p->seq == 0)
	{
		rx_missed = 0;
		tx_seq_count = 0;
	}
	else if (rx_count != 0 && (rx_last_id + 1) != p->seq)
	{
		rx_missed += p->seq - rx_last_id - 1;
	}

	rx_last_id = p->seq;
	rx_count++;

	send_reply(cpu_id, rmq, p->payload_len);

	mq_discard(TRTL_RMQ, rmq);
}

uint32_t last_ts = 0;

void print_stats()
{
	uint32_t ts = trtl_get_runtime_milliseconds();

	if ((last_ts == 0) || (ts >= (last_ts + 1000)))
	{
		pp_printf("Stats: RX %"PRId32", missed %"PRId32", errors %"PRId32"\n\r",
			  rx_count, rx_missed, rx_errors);
		last_ts = ts;
	}
}

int main()
{
	int cpu = trtl_get_core_id();
	int rmq = 0;

	last_ts = 0;

	struct trtl_ep_address bind_addr;

	pp_printf("TEST for: RMQ UDP send/receive, slot %d, cpu %d\r\n", rmq, cpu);

	// set up the RMQ Endpoint

	// we operate only with UDP frames
	bind_addr.type = TRTL_EP_FRAME_UDP;

	// destination MAC: we use broadcast
	bind_addr.dst_mac[0] = 0xff;
	bind_addr.dst_mac[1] = 0xff;
	bind_addr.dst_mac[2] = 0xff;
	bind_addr.dst_mac[3] = 0xff;
	bind_addr.dst_mac[4] = 0xff;
	bind_addr.dst_mac[5] = 0xff;
	
	// destination port
	bind_addr.dst_port = 12345;
	// source port
	bind_addr.src_port = 7777;
	// destination IP: 192.168.90.255 (broadcast)
	bind_addr.dst_ip = 0xC0A85AFF;
	// source IP: 192.168.90.255 (broadcast)
	bind_addr.src_ip = 0xC0A85A11;
	bind_addr.ethertype = 0x800; // IPv4
	// RX filter: we want only UDP packets with matching desination port & IP address.
	bind_addr.filter = TRTL_EP_FILTER_UDP | TRTL_EP_FILTER_DST_PORT | TRTL_EP_FILTER_DST_IP;

	// configure outgoing channel
	mq_bind(TRTL_RMQ, rmq, 1, &bind_addr);
	// configure incoming channel
	mq_bind(TRTL_RMQ, rmq, 0, &bind_addr);

	for (;;)
	{
		handle_rx(cpu, rmq);
		print_stats();
	}

	return 0;
}
